/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.cdt.internal.ui.cview;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.cdt.core.CCorePlugin;
import org.eclipse.cdt.core.ProcessorContainerChangeEvent;
import org.eclipse.cdt.core.model.ICElement;
import org.eclipse.cdt.core.model.ICProject;
import org.eclipse.cdt.internal.core.model.CModelManager;
import org.eclipse.cdt.internal.core.model.ProcessorContainer;
import org.eclipse.cdt.internal.ui.ICHelpContextIds;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.actions.RenameResourceAction;


/**
 * The ResourceNavigatorRenameAction is the rename action used by the
 * ResourceNavigator that also allows updating after rename.
 * @since 2.0
 */
public class TreeCViewRenameAction extends RenameResourceAction {
	private TreeViewer viewer;
	/**
	 * Create a ResourceNavigatorRenameAction and use the tree of the supplied viewer
	 * for editing.
	 * @param shell Shell
	 * @param treeViewer TreeViewer
	 */
	public TreeCViewRenameAction(Shell shell, TreeViewer treeViewer) {
		super(shell, treeViewer.getTree());
		PlatformUI.getWorkbench().getHelpSystem().setHelp(
			this,
			ICHelpContextIds.RENAME_ACTION);
		this.viewer = treeViewer;
	}
	/* (non-Javadoc)
	 * Run the action to completion using the supplied path.
	 */
	protected void runWithNewPath(IPath path, IResource resource) {
		IWorkspaceRoot root = resource.getProject().getWorkspace().getRoot();
		super.runWithNewPath(path, resource);
		if (this.viewer != null) {
			IResource newResource = root.findMember(path);
			if (newResource != null) {
				this.viewer.setSelection(new StructuredSelection(newResource), true);
				if (resource.getType() == IResource.PROJECT) {
					updateProjectContainer(resource.getProject(), newResource.getProject());
				}
			}
		}
	}
	/**
	* Handle the key release
	*/
	public void handleKeyReleased(KeyEvent event) {
		if (event.keyCode == SWT.F2 && event.stateMask == 0 && isEnabled()) {
			run();
		}
	}
	
	/**
	 * Removes the old project from its parent ProcessorContainer.  If the renamed project is a Software
	 * Platform project, update all application projects which reference it.
	 * 
	 * @param oldProj
	 * @param newProj
	 */
	private void updateProjectContainer(IProject oldProj, IProject newProj) {
//		RootElement root = (RootElement)viewer.getInput();
		String oldProjName = oldProj.getName();
		
		try {			
			IProjectDescription desc = newProj.getDescription();
			boolean isSwPlatform = newProj.hasNature("com.xilinx.mdt.edk.swPlatformProject");
			ProcessorContainer pContainer = CModelManager.getDefault().getProcessorContainer(desc.getComment());
			if (pContainer != null) {
				List list = pContainer.getChildrenOfType(ICElement.C_PROJECT);
				ICProject cproj;
				IProject proj;
				for (Iterator it = list.iterator(); it.hasNext();) {
					cproj = (ICProject)it.next();
					if (cproj.getElementName().equals(oldProjName/*oldProj.getName()*/)) {
						pContainer.removeChild(cproj);
					} else {
						proj = cproj.getProject();
						if (isSwPlatform) {
							if (!proj.hasNature("com.xilinx.mdt.edk.swPlatformProject") 
									&& !proj.hasNature("com.xilinx.mdt.edk.bspProject")) {
								desc = proj.getDescription();
								IProject[] refProjs = desc.getReferencedProjects();
								IProject[] newRefProjs = new IProject[refProjs.length];
								for (int i = 0; i < refProjs.length; i++) {
									if (refProjs[i].getName().equals(oldProj.getName())) {
										newRefProjs[i] = newProj;
									} else {
										newRefProjs[i] = refProjs[i];
									}
								}
								if (refProjs.length > 0) {
									desc.setReferencedProjects(newRefProjs);
									proj.setDescription(desc, null);
								}
							}
						}
						if (proj.hasNature("com.xilinx.mdt.edk.swPlatformProject")
								|| proj.hasNature("com.xilinx.mdt.edk.bspProject")) {
							// rename mss file
							new File(proj.getLocation().toString() + "/" + oldProj.getName() + ".mss").renameTo(
									new File(proj.getLocation().toString() + "/" + proj.getName() + ".mss"));
							updateMakefile(proj.getLocation().toString() + "/" + "libgen.mk", proj.getName() + ".mss");
						}
					}
				}
				CCorePlugin.getDefault().fireProcessorContainerChanged(new ProcessorContainerChangeEvent(pContainer.getElementName()));
			}			
		} catch(CoreException ex) {
			//System.out.println("updateProjectContainer: caught exception: " + ex.getMessage());
		}
	}
	
	/**
	 * Changes MSSFILE in libgen.mk to the new mss file name.
	 * 
	 * @param makefile
	 * @param mssFile
	 */
	private void updateMakefile(String makefile, String mssFile) {
		try {
			List list = new ArrayList();
			BufferedReader reader = new BufferedReader(new FileReader(makefile));
	        String line = reader.readLine();
	        while (line != null) {
	        	list.add(line);
	        	line = reader.readLine();
	        }
	        reader.close();
	        int idx;
	        List newList = new ArrayList();
	        for (Iterator it = list.iterator(); it.hasNext();) {
	        	line = (String)it.next();
	        	idx = line.indexOf("MSSFILE =");
	        	if (idx != -1) {
	        		line = "MSSFILE = " + mssFile;
	        	}
	        	newList.add(line);
	        }
	        PrintWriter pw = new PrintWriter(new FileWriter(makefile, false));
	        for (Iterator it = newList.iterator(); it.hasNext();) {
	        	pw.println((String)it.next());
	        }
	        pw.close();
		} catch(IOException ex) {
			//System.out.println("updateMakefile: " + ex.getMessage());
		}
	}
}
